##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################
__doc__='''Stored Procedure

$Id: SP.py,v 1.8 2001/10/09 18:05:02 matt Exp $'''
__version__='$Revision: 1.8 $'[11:-2]

import sys
from Globals import HTMLFile
from OFS.SimpleItem import SimpleItem
from Shared.DC.ZRDB.TM import TM
from string import split
from AccessControl import getSecurityManager
import DCOracle2

myname = split(__name__,".")[1]
if myname == "ZOracleDA":
    dbstring = "Z Oracle"
else:
    dbstring = myname

class Procedure(SimpleItem, TM):

    manage_options=(
        {'label': 'Edit', 'action':'manage_procview'},
        {'label': 'Test', 'action':'manage_proctest'},
        # {'label': 'Design', 'action':'manage_tables'},
        ) + SimpleItem.manage_options

    __ac_permissions__ =  (
        ('Change %s Stored Procedures' % dbstring, 
         ('manage_editZOracleStoredProcedure',)),
        ('Execute %s Stored Procedures' % dbstring,
         ('manage_testZOracleStoredProcedure',"")),
    )

    manage_procview=HTMLFile('dtml/procview',globals())
    manage_proctest=HTMLFile('dtml/proctest',globals())

    info=None
        
    def __init__(self, id, title, connection, procname, check):
        self.id = id
        self.title = title
        self.connection = connection
        self.procname = procname
        self._v_proc = None
        self._v_db = None

    
    def _connect(self):
        db = self._v_db = getattr(self,self.connection)() # Get connection
        c = db.cursor                          # Get cursor
        p = c.findproc(self.procname)        # Get procedure
        self._v_proc = p

    def _finish(self, *ignored):
        self._v_db.db.commit()

    def _abort(self, *ignored):
        self._v_db.db.rollback()

    def description(self):
        if getattr(self,"_v_proc",None) is None: self._connect()
        return self._v_proc.__doc__

    def __call__(self, *args, **kw):
        if not getSecurityManager().checkPermission(
            'Execute %s Stored Procedures' % dbstring, self):
            # Ugh raising a string error
            raise 'Unauthorized', "You are not authorized to access this resource"
        self._register()
        if getattr(self,"_v_proc",None) is None: self._connect()
        try:
            # Note, this does not do result promotion like the DA query will
            # which is probably bad.  OracleDates in particular look like
            # DateTimes but arent at all the same!
            return apply(self._v_proc,args,kw)
        except:
            self._v_proc = None             
            raise                          # Reraise error

    def invars(self):
        d = self.description()
        ins = []
        lines = split(d,'\n')
        if len(lines) > 1:
            for l in lines[1:]:
                words = split(l)
                if len(words) > 1 and words[1] == 'IN': ins.append(words[0])
        return ins

    def manage_editZOracleStoredProcedure(self, title, connection, procname,
        URL1, RESPONSE=None):
        """Edit the stored procedure"""
        self.title = title
        self.connection = connection
        self.procname = procname
        self._v_proc = None
        RESPONSE.redirect(URL1 + "/manage_procview?manage_tabs_message=Stored%20Procedure%20Changed")

    def manage_testZOracleStoredProcedure(self, REQUEST):
        """Test the stored procedure"""
        v = apply(self,[],REQUEST.form)
        return v
